/*
GNU General Public License

Nail6 - Java Image Thumbnail & Index Creation Tool
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.nail6;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreeSelectionModel;

import javax.media.jai.BorderExtender;
import javax.media.jai.Interpolation;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.RenderedOp;

import com.sun.media.jai.codec.FileSeekableStream;
import com.sun.media.jai.codec.ImageCodec;
import com.sun.media.jai.codec.ImageDecoder;
import com.sun.media.jai.codec.SeekableStream;

import com.hexidec.nail6.NailIcon;
import com.hexidec.util.TreeSpider;

/** Nail6
  * App for creating thumbnail pages
  *
  * @author Howard Kistler
  * @version 0.1
  *
  * REQUIREMENTS
  * Java 2 (JDK 1.3)
  * Swing
  * Java Advanced Imaging library
  *
  * VERSION HISTORY
  * 0.1 (09/14/2000) - initial creation (08/09/2000)
  *
  * FORTHCOMING FEATURES & FIXES
  * - Set default application start-up size
  * - Specify the directory search depth
  * - Export image as Windows ICO file
  * - "Live" changes (e.g., changing default nail size updates relevant menu item)
  */

public class Nail6 extends JFrame implements WindowListener, ActionListener, TreeSelectionListener
{
	// Application Components
	private JScrollPane jscpViewer;
	private JTree jTree;
	private JScrollPane jscpTree;
	private JSplitPane jSplitPane;
	private JLabel jlblStatusBar;

	// Menu Components
	private JMenuBar jMenuBar;
	private JMenu jMenuActions;
	private JMenuItem jmiActnsOpen;
	private JMenuItem jmiActnsCDir;
	private JMenuItem jmiActnsScan;
	private JMenuItem jmiActnsWriteTree;
	private JMenuItem jmiActnsWriteIndex;
	private JMenuItem jmiActnsExit;
	private JMenu jMenuImages;
	private JCheckBoxMenuItem[] jmiImageTypes;
	private JMenuItem jmiImageGIF;
	private JMenuItem jmiImageJPG;
	private JMenuItem jmiImagePNG;
	private JMenuItem jmiImageTIF;
	private JMenuItem jmiImageBMP;
	private JMenu jMenuNailFormats;
	private JRadioButtonMenuItem[] jmiNailFormats;
	private JMenu jMenuNailSize;
	private JRadioButtonMenuItem[] jmiNailSizes;
	private JMenu jMenuAntialias;
	private JRadioButtonMenuItem[] jmiInterpTypes;
	private JMenu jMenuOptions;
	private JMenuItem jmiSetOptions;
	private JCheckBoxMenuItem jmiViewAsNail;
	private JCheckBoxMenuItem jmiReplaceNails;

	// Local TreeSpider instance for file tree traversal and searching
	private TreeSpider spider;

	// Options Collections
	private Vector vcImageTypes;
	private Vector vcNailFormats;
	private Vector vcNailSizes;
	private Vector vcInterpTypes;
	private Vector vcExts;
	private String[] strExts = { };
	private String strDesc = "";

	// Scaled PlanarImage RenderingHints
	private RenderingHints rHints = new RenderingHints(JAI.KEY_BORDER_EXTENDER , BorderExtender.createInstance(BorderExtender.BORDER_COPY));

	// Image Display Components
	private Image image = (Image)null;
	private JLabel jlblIconDisplay;
	private Icon iconImage;

	// Application Properties Table
	private Hashtable htPropertyTable = new Hashtable();

	// Sequential Index Of Application Properties
	private Vector vcPropertySequence = new Vector();

	// Application Properties Dialog
	private AppOptionsDialog appOpDialog;

	// Base Application Settings
	private String baseDir;
	private String initFilename;

	// String Constants
	private static final String APPTITLE        = "Nail6";
	private static final String MSG_APPREADY    = "Ready";
	private static final String ARRAY_IMAGETYPE = "[IMGTYPE]";

// Constructors ----------------------------------------------------------->

	/* Main Constructor */
	public Nail6(String baseDirName, String initFile)
	{
		/* Initialise app */
		try
		{
			baseDir = new File(baseDirName).getCanonicalPath();
		}
		catch(IOException ioe)
		{
			sysExit("Error getting default canonical path : " + ioe.getMessage(), 1);
		}

		/* Read in init file */
		initFilename = initFile;
		readInitSettings();

		/* Instantiate an appropriate TreeSpider to perform file handling */
		if(appPropertyExists("DEFAULT_IMAGE_DIR"))
		{
			baseDir = getAppPropertyAsString("DEFAULT_IMAGE_DIR");
		}
		spider = new TreeSpider(baseDir, false, 0);

		/* Initialise supported image types */
		vcImageTypes = new Vector();
		vcImageTypes.add(new CustomKeyValuePair("GIF",  new String[] { ".gif" }));
		vcImageTypes.add(new CustomKeyValuePair("JPEG", new String[] { ".jpeg", ".jpg" }));
		vcImageTypes.add(new CustomKeyValuePair("PNG",  new String[] { ".png" }));
		vcImageTypes.add(new CustomKeyValuePair("TIFF", new String[] { ".tiff", ".tif" }));
		vcImageTypes.add(new CustomKeyValuePair("BMP",  new String[] { ".bmp" }));
		vcImageTypes.add(new CustomKeyValuePair("PNM",  new String[] { ".pnm", ".pbm", ".pgm", ".ppm" }));
		vcImageTypes.add(new CustomKeyValuePair("FPX",  new String[] { ".fpx" }));

		/* Initialise nail formats */
		vcNailFormats = new Vector();
		vcNailFormats.add(new CustomKeyValuePair("JPEG", new String[] { "jpeg" }));
		vcNailFormats.add(new CustomKeyValuePair("PNG",  new String[] { "png" }));

		/* Initialise nail sizes */
		vcNailSizes = new Vector();
		vcNailSizes.add(new CustomKeyValuePair("Tiny",   new String[] { "16" }));
		vcNailSizes.add(new CustomKeyValuePair("Small",  new String[] { "32" }));
		vcNailSizes.add(new CustomKeyValuePair("Medium", new String[] { "48" }));
		vcNailSizes.add(new CustomKeyValuePair("Large",  new String[] { "64" }));
		vcNailSizes.add(new CustomKeyValuePair("Huge",   new String[] { "128" }));
		vcNailSizes.add(new CustomKeyValuePair("Custom", new String[] { getAppPropertyAsString("CUSTOM_NAIL_SIZE") }));

		/* Initialise interpolation types */
		vcInterpTypes = new Vector();
		vcInterpTypes.add(new CustomKeyValuePair("Nearest Neighbor", new Object[] { Interpolation.getInstance(Interpolation.INTERP_NEAREST) } ));
		vcInterpTypes.add(new CustomKeyValuePair("Bilinear",         new Object[] { Interpolation.getInstance(Interpolation.INTERP_BILINEAR) } ));
		vcInterpTypes.add(new CustomKeyValuePair("Bicubic",          new Object[] { Interpolation.getInstance(Interpolation.INTERP_BICUBIC) } ));
		vcInterpTypes.add(new CustomKeyValuePair("Bicubic 2",        new Object[] { Interpolation.getInstance(Interpolation.INTERP_BICUBIC_2) } ));

		/* Set up the menus */
		jmiImageTypes  = new JCheckBoxMenuItem[vcImageTypes.size()];
		jmiNailFormats = new JRadioButtonMenuItem[vcNailFormats.size()];
		jmiNailSizes   = new JRadioButtonMenuItem[vcNailSizes.size()];
		jmiInterpTypes = new JRadioButtonMenuItem[vcInterpTypes.size()];

		jMenuActions = new JMenu("Actions");
		jmiActnsOpen       = new JMenuItem("Open Image");        jmiActnsOpen.setActionCommand("open");             jmiActnsOpen.addActionListener(this);       jMenuActions.add(jmiActnsOpen);
		jmiActnsCDir       = new JMenuItem("Change Directory");  jmiActnsCDir.setActionCommand("chdir");            jmiActnsCDir.addActionListener(this);       jMenuActions.add(jmiActnsCDir);
		jmiActnsScan       = new JMenuItem("Rescan For Images"); jmiActnsScan.setActionCommand("scan");             jmiActnsScan.addActionListener(this);       jMenuActions.add(jmiActnsScan);
		jmiActnsWriteTree  = new JMenuItem("Write Tree Page");   jmiActnsWriteTree.setActionCommand("writetree");   jmiActnsWriteTree.addActionListener(this);  jMenuActions.add(jmiActnsWriteTree);
		jmiActnsWriteIndex = new JMenuItem("Write Index Page");  jmiActnsWriteIndex.setActionCommand("writeindex"); jmiActnsWriteIndex.addActionListener(this); jMenuActions.add(jmiActnsWriteIndex);
		jmiActnsExit       = new JMenuItem("Exit");              jmiActnsExit.setActionCommand("exit");             jmiActnsExit.addActionListener(this);       jMenuActions.add(jmiActnsExit);

		jMenuImages = new JMenu("Image Types");
		for(int i = 0; i < vcImageTypes.size(); i++)
		{
			String strImageType = ((CustomKeyValuePair)(vcImageTypes.elementAt(i))).getCustomKey();
			jmiImageTypes[i] = new JCheckBoxMenuItem(strImageType, appPropertyExists(ARRAY_IMAGETYPE + strImageType));
			jmiImageTypes[i].setActionCommand("imagetypes");
			jmiImageTypes[i].addActionListener(this);
			jMenuImages.add(jmiImageTypes[i]);
		}

		jMenuNailFormats = new JMenu("Nail Format");
		ButtonGroup btngrpNailFormats = new ButtonGroup();
		for(int i = 0; i < vcNailFormats.size(); i++)
		{
			jmiNailFormats[i] = new JRadioButtonMenuItem(((CustomKeyValuePair)(vcNailFormats.elementAt(i))).getCustomKey(), (i == 0));
			btngrpNailFormats.add(jmiNailFormats[i]);
			jMenuNailFormats.add(jmiNailFormats[i]);
		}

		jMenuNailSize = new JMenu("Nail Size");
		ButtonGroup btngrpNailSize = new ButtonGroup();
		for(int i = 0; i < vcNailSizes.size(); i++)
		{
			CustomKeyValuePair ckvpNailSize = (CustomKeyValuePair)(vcNailSizes.elementAt(i));
			String strNailLabel = (ckvpNailSize.getCustomKey() + " (" + ckvpNailSize.getCustomValueAsInteger(0) + "px)");
			jmiNailSizes[i] = new JRadioButtonMenuItem(strNailLabel, (ckvpNailSize.getCustomValueAsInteger(0) == getAppPropertyAsInteger("DEFAULT_NAIL_SIZE")));
			btngrpNailSize.add(jmiNailSizes[i]);
			jMenuNailSize.add(jmiNailSizes[i]);
		}

		jMenuAntialias = new JMenu("Antialiasing");
		ButtonGroup btngrpAntialias = new ButtonGroup();
		for(int i = 0; i < vcInterpTypes.size(); i++)
		{
			String strAntiLabel = ((CustomKeyValuePair)(vcInterpTypes.elementAt(i))).getCustomKey();
			jmiInterpTypes[i] = new JRadioButtonMenuItem(strAntiLabel, strAntiLabel.equals(getAppPropertyAsString("DEFAULT_ANTIALIAS")));
			btngrpAntialias.add(jmiInterpTypes[i]);
			jMenuAntialias.add(jmiInterpTypes[i]);
		}

		jMenuOptions = new JMenu("Options");
		jmiSetOptions   = new JMenuItem("Set Options");                                                   jmiSetOptions.setActionCommand("setoptions"); jmiSetOptions.addActionListener(this); jMenuOptions.add(jmiSetOptions);
		jmiViewAsNail   = new JCheckBoxMenuItem("View As Nail", getAppPropertyAsBoolean("VIEW_AS_NAIL")); jMenuOptions.add(jmiViewAsNail);
		jmiReplaceNails = new JCheckBoxMenuItem("Replace Existing Nails", false);                         jMenuOptions.add(jmiReplaceNails);

		jMenuBar = new JMenuBar();
		jMenuBar.add(jMenuActions);
		jMenuBar.add(jMenuImages);
		jMenuBar.add(jMenuNailFormats);
		jMenuBar.add(jMenuNailSize);
		jMenuBar.add(jMenuAntialias);
		jMenuBar.add(jMenuOptions);

		/* Build the current accepted file extensions listing */
		vcExts = new Vector();
		buildExtsSettings();

		/* Set up the JTree and JScrollPane for exploring the directory structure */
		jTree = new JTree(new javax.swing.tree.DefaultMutableTreeNode("[root]"));
		jTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		jTree.addTreeSelectionListener(this);
		jscpTree = new JScrollPane(jTree);
		jscpTree.setPreferredSize(new Dimension(200,400));

		/* Set up the JLabel and JScrollPane for viewing selected images */
		jlblIconDisplay = new JLabel();
		jlblIconDisplay.setPreferredSize(new Dimension(400,400));
		jscpViewer = new JScrollPane(jlblIconDisplay);

		/* Add file tree and image viewer to JSplitPane */
		jSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, jscpTree, jscpViewer);

		/* Create the status bar */
		jlblStatusBar = new JLabel("Ready");

		/* Set up the main window */
		this.addWindowListener(this);
		this.setJMenuBar(jMenuBar);
		this.getContentPane().setLayout(new BorderLayout());
		this.getContentPane().add(jSplitPane, BorderLayout.CENTER);
		this.getContentPane().add(jlblStatusBar, BorderLayout.SOUTH);

		/* Show the main app window */
		this.pack();
		this.show();

		/* Bring app into ready state */
		setTitleBarText(baseDir);
		setStatusBarText("Scanning " + baseDir + " for images...");
		scanForImages();
		setStatusBarText(MSG_APPREADY);
	}

	/* Alternate constructor */
	public Nail6(String baseDirName)
	{
		this(baseDirName, "nail6.ini");
	}

// Initialisation Methods ------------------------------------------------->

	/** Reads initialisation file and sets appropriate values */
	private void readInitSettings()
	{
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(initFilename));
			String lineIn = "";
			do
			{
				lineIn = br.readLine();
				if(lineIn == null)
				{
					break;
				}
				else if(lineIn.startsWith("!"))
				{
					/* this is a comment line */;
				}
				else if(lineIn.startsWith("["))
				{
					/* this is an array member */;
					if(lineIn.indexOf(ARRAY_IMAGETYPE) > -1)
					{
						if(valueFromInit(lineIn).equalsIgnoreCase("true"))
						{
							htPropertyTable.put(propFromInit(lineIn), "");
						}
					}
				}
				else if(lineIn.indexOf("=") > -1)
				{
					/* this is an application property */
					htPropertyTable.put(propFromInit(lineIn), valueFromInit(lineIn));
					vcPropertySequence.add(propFromInit(lineIn));
				}
				else
				{
					System.out.println("Ignoring unknown entry type in init file : " + lineIn);
				}
			} while (lineIn != null);
			br.close();
		}
		catch(IOException ioe)
		{
			sysExit("Error initialising from init file (" + initFilename + ") : "  + ioe.getMessage(), 2);
		}
	}

	/** Writes initialisation file */
	private void writeInitSettings(Vector vcImageClasses)
	{
		Vector vcFileLines = new Vector();
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(initFilename));
			String lineIn = "";
			do
			{
				/* read in the comments, which will always filter up to the top of the file on write out */
				lineIn = br.readLine();
				if     (lineIn == null)         { break; }
				else if(lineIn.startsWith("!")) { vcFileLines.add(new String(lineIn)); }
				else                            { /* ignore */; }
			} while (lineIn != null);
			br.close();
			/* write out the application properties sequentially */
			for(int i = 0; i < vcPropertySequence.size(); i++)
			{
				String propName = (String)(vcPropertySequence.elementAt(i));
				vcFileLines.add(propName + "=" + getAppPropertyAsString(propName));
			}
			/* write out the image type settings */
			for(int i = 0; i < vcImageClasses.size(); i++)
			{
				vcFileLines.add((String)(vcImageClasses.elementAt(i)));
			}
			BufferedWriter bw = new BufferedWriter(new FileWriter(initFilename));
			for(int i = 0; i < vcFileLines.size(); i++)
			{
				String lineOut = (String)(vcFileLines.elementAt(i));
				bw.write(lineOut, 0, lineOut.length());
				bw.newLine();
			}
			bw.flush();
			bw.close();
		}
		catch(IOException ioe)
		{
			sysExit("Error writing to init file (" + initFilename + ") : "  + ioe.getMessage(), 3);
		}
	}

// Listener Methods ------------------------------------------------------->

	/* WindowListener methods */
	public void windowClosing(WindowEvent we)
	{
		this.dispose();
		System.exit(0);
	}
	public void windowOpened(WindowEvent we)      { ; }
	public void windowClosed(WindowEvent we)      { ; }
	public void windowActivated(WindowEvent we)   { ; }
	public void windowDeactivated(WindowEvent we) { ; }
	public void windowIconified(WindowEvent we)   { ; }
	public void windowDeiconified(WindowEvent we) { ; }

	/* ActionListener methods */
	public void actionPerformed(ActionEvent ae)
	{
		if(ae.getActionCommand().equals("exit"))
		{
			this.dispose();
			System.exit(0);
		}
		else if(ae.getActionCommand().equals("open"))
		{
			setStatusBarText("Opening file...");
			File whatFile = getFile(JFileChooser.OPEN_DIALOG, JFileChooser.FILES_AND_DIRECTORIES, strExts, strDesc);
			if(whatFile != null)
			{
				showSelectedImage(whatFile);
				setTitleBarText(whatFile.getName());
			}
			setStatusBarText(MSG_APPREADY);
		}
		else if(ae.getActionCommand().equals("chdir"))
		{
			setStatusBarText("Changing directory...");
			File whatFile = getFile(JFileChooser.OPEN_DIALOG, JFileChooser.DIRECTORIES_ONLY, strExts, strDesc);
			if(whatFile != null)
			{
				baseDir = whatFile.getPath();
				scanForImages();
				setTitleBarText(baseDir);
			}
			setStatusBarText(MSG_APPREADY);
		}
		else if(ae.getActionCommand().equals("scan"))
		{
			setStatusBarText("Scanning " + baseDir + " for images...");
			scanForImages();
			setStatusBarText(MSG_APPREADY);
		}
		else if(ae.getActionCommand().equals("writetree"))
		{
			File whatFile = getFile(JFileChooser.SAVE_DIALOG, JFileChooser.FILES_AND_DIRECTORIES, new String[] { ".html", ".htm" }, "HTML Files (*.html, *.htm)");
			if(whatFile != null)
			{
				setStatusBarText("Writing tree to file " + whatFile.getPath() + "...");
				writeTreeFileListing(whatFile);
				setStatusBarText(MSG_APPREADY);
			}
		}
		else if(ae.getActionCommand().equals("writeindex"))
		{
			File whatFile = getFile(JFileChooser.SAVE_DIALOG, JFileChooser.FILES_AND_DIRECTORIES, new String[] { ".html", ".htm" }, "HTML Files (*.html, *.htm)");
			if(whatFile != null)
			{
				setStatusBarText("Writing index to file " + whatFile.getPath() + "...");
				writeIndexFileListing(whatFile);
				setStatusBarText(MSG_APPREADY);
			}
		}
		else if(ae.getActionCommand().equals("imagetypes"))
		{
			buildExtsSettings();
			scanForImages();
		}
		else if(ae.getActionCommand().equals("setoptions"))
		{
			this.repaint();
			appOpDialog = new AppOptionsDialog(this);
		}
		else if(ae.getActionCommand().equals("closeoptions"))
		{
			if(appOpDialog != null)
			{
				appOpDialog.hide();
				appOpDialog.dispose();
				appOpDialog = null;
			}
			this.repaint();
		}
		else if(ae.getActionCommand().equals("saveoptions"))
		{
			if(appOpDialog != null)
			{
				for(int i = 0; i < vcPropertySequence.size(); i++)
				{
					String appProp = (String)(vcPropertySequence.elementAt(i));
					if(appProp.equals("LINK_TO_SOURCE") || appProp.equals("NICE_FORMATTING") || appProp.equals("DEFAULT_NAIL_SIZE") || appProp.equals("DEFAULT_ANTIALIAS"))
					{
						// ignore check boxes & radio buttons here
					}
					else
					{
						updateAppProperty(appProp, appOpDialog.jtxtAppProperties[i].getText());
					}
				}
				updateAppProperty("LINK_TO_SOURCE",   "" + appOpDialog.jchkLINK_TO_SOURCE.isSelected() + "");
				updateAppProperty("NICE_FORMATTING", "" + appOpDialog.jchkNICE_FORMATTING.isSelected() + "");
				updateAppProperty("VIEW_AS_NAIL",     "" + appOpDialog.jchkVIEW_AS_NAIL.isSelected() + "");
				for(int i = 0; i < appOpDialog.jrbtNailSizes.length; i++)
				{
					if(appOpDialog.jrbtNailSizes[i].isSelected())
					{
						updateAppProperty("DEFAULT_NAIL_SIZE", ((CustomKeyValuePair)(vcNailSizes.elementAt(i))).getCustomValueAsString(0));
						break;
					}
				}
				for(int i = 0; i < appOpDialog.jrbtAntiAlias.length; i++)
				{
					if(appOpDialog.jrbtAntiAlias[i].isSelected())
					{
						updateAppProperty("DEFAULT_ANTIALIAS", ((CustomKeyValuePair)(vcInterpTypes.elementAt(i))).getCustomKey());
						break;
					}
				}
				Vector vcImgTypes = new Vector();
				for(int i = 0; i < appOpDialog.jchkImageTypes.length; i++)
				{
					if(appOpDialog.jchkImageTypes[i].isSelected())
					{
						updateAppProperty(ARRAY_IMAGETYPE + appOpDialog.jchkImageTypes[i].getText(), "");
					}
					else
					{
						removeAppProperty(ARRAY_IMAGETYPE + appOpDialog.jchkImageTypes[i].getText());
					}
					vcImgTypes.add(new String(ARRAY_IMAGETYPE + appOpDialog.jchkImageTypes[i].getText() + "=" + appOpDialog.jchkImageTypes[i].isSelected()));
				}
				writeInitSettings(vcImgTypes);
				appOpDialog.hide();
				appOpDialog.dispose();
				appOpDialog = null;
			}
			this.repaint();
		}
		else if(ae.getActionCommand().equals("browseforimagedir"))
		{
			File whatFile = getFile(JFileChooser.OPEN_DIALOG, JFileChooser.DIRECTORIES_ONLY, null, null);
			if(whatFile != null)
			{
				int whichTextField = -1;
				for(int i = 0; i < vcPropertySequence.size(); i++)
				{
					if(((String)(vcPropertySequence.elementAt(i))).equals("DEFAULT_IMAGE_DIR"))
					{
						whichTextField = i;
						break;
					}
				}
				if(whichTextField > -1)
				{
					appOpDialog.jtxtAppProperties[whichTextField].setText(whatFile.getPath());
				}
			}
			appOpDialog.requestFocus();
		}
		else if(ae.getActionCommand().equals("browsefornaildir"))
		{
			File whatFile = getFile(JFileChooser.OPEN_DIALOG, JFileChooser.DIRECTORIES_ONLY, null, null);
			if(whatFile != null)
			{
				int whichTextField = -1;
				for(int i = 0; i < vcPropertySequence.size(); i++)
				{
					if(((String)(vcPropertySequence.elementAt(i))).equals("NAIL_DIR"))
					{
						whichTextField = i;
						break;
					}
				}
				if(whichTextField > -1)
				{
					appOpDialog.jtxtAppProperties[whichTextField].setText(whatFile.getPath());
				}
			}
			appOpDialog.requestFocus();
		}
		else
		{
			System.out.println("ActionEvent : " + ae);
		}
	}

	/* TreeSelectionListener method */
	public void valueChanged(TreeSelectionEvent tse)
	{
		if(jTree.getSelectionCount() > 0)
		{
			StringBuffer sbPath = new StringBuffer();
			Object[] treePath = jTree.getSelectionPath().getPath();
			for(int i = 1; i < treePath.length; i++)
			{
				if(i > 1)
				{
					sbPath.append(File.separator);
				}
				sbPath.append(treePath[i]);
			}
			File selFile = new File(baseDir + File.separator + sbPath.toString());
			if(!selFile.isDirectory())
			{
				showSelectedImage(selFile);
			}
		}
	}

// Class Methods ---------------------------------------------------------->

	/* Builds a listing of acceptable file extensions */
	private void buildExtsSettings()
	{
		vcExts.removeAllElements();
		StringBuffer strBuffer = new StringBuffer();
		strBuffer.append("Supported Formats (");
		for(int i = 0; i < vcImageTypes.size(); i++)
		{
			if(jmiImageTypes[i].isSelected())
			{
				String[] exts = ((CustomKeyValuePair)(vcImageTypes.elementAt(i))).getCustomValueAsStringArray();
				for(int j = 0; j < exts.length; j++)
				{
					vcExts.add(exts[j]);
					strBuffer.append((((i > 0) || (j > 0)) ? "," : "") + "*" + exts[j]);
				}
			}
		}
		strBuffer.append(")");
		strDesc = strBuffer.toString();
		strExts = new String[vcExts.size()];
		for(int i = 0; i < vcExts.size(); i++)
		{
			strExts[i] = (String)(vcExts.elementAt(i));
		}
	}

	/* Scans the current directory tree for accepted image file types */
	private void scanForImages()
	{
		spider.setRootFile(new File(baseDir));
		spider.setExtensions(vcExts);
		jTree.setModel(spider.fetchTree());
	}

	/* Display the selected image in the image viewer panel */
	private void showSelectedImage(File imageFile)
	{
		try
		{
			PlanarImage pimgThumbnail;
			if(jmiViewAsNail.isSelected())
			{
				pimgThumbnail = getScaledPlanarImage(imageFile.getPath());
			}
			else
			{
				pimgThumbnail = getPlanarImage(imageFile.getPath());
			}
			iconImage = new NailIcon(pimgThumbnail);
	        jlblIconDisplay.setIcon(iconImage);
	        Dimension iconSize = new Dimension(iconImage.getIconWidth(), iconImage.getIconHeight());
	        jlblIconDisplay.setPreferredSize(iconSize);
	        jlblIconDisplay.revalidate();
		}
		catch(Exception e)
		{
			System.out.println("Exception occurred in showSelectedImage() on file " + imageFile.getPath());
		}
	}

	/* Load and scale the selected image */
	private PlanarImage getScaledPlanarImage(String imageFile)
	{
		/* determine Nail Size from menus */
		int nailSize = getAppPropertyAsInteger("CUSTOM_NAIL_SIZE");
		for(int i = 0; i < jmiNailSizes.length; i++)
		{
			if(jmiNailSizes[i].isSelected())
			{
				nailSize = ((CustomKeyValuePair)(vcNailSizes.elementAt(i))).getCustomValueAsInteger(0);
			}
		}
		/* determine Interpolation from menus */
		Interpolation interpol = Interpolation.getInstance(Interpolation.INTERP_NEAREST);
		for(int i = 0; i < jmiInterpTypes.length; i++)
		{
			if(jmiInterpTypes[i].isSelected())
			{
				interpol = (Interpolation)(((CustomKeyValuePair)(vcInterpTypes.elementAt(i))).getCustomValueAsObject(0));
			}
		}
		/* get scaled instance of image */
		try
		{
			PlanarImage pimgThumbnail = scaleImage(getPlanarImage(imageFile), nailSize, interpol);
			return pimgThumbnail;
		}
		catch(Exception e)
		{
			System.out.println("Exception occurred in getScaledPlanarImage() on file " + imageFile);
		}
		return (PlanarImage)null;
	}

	/** Retrieves an image as a PlanarImage
	  */
	public PlanarImage getPlanarImage(String imagename)
	{
        if (System.getProperty("JAI_IMAGE_READER_USE_CODECS") == null)
        {
            return JAI.create("fileload", imagename);
        }
        else
        {
            try
            {
                SeekableStream stream = new FileSeekableStream(imagename);
                String[] names = ImageCodec.getDecoderNames(stream);
                ImageDecoder dec = ImageCodec.createImageDecoder(names[0], stream, null);
                RenderedImage im = dec.decodeAsRenderedImage();
                return PlanarImage.wrapRenderedImage(im);
            }
            catch(Exception e)
            {
                e.printStackTrace();
                return null;
            }
        }
    }

	/** Scales and interpolates an image
	  */
    public PlanarImage scaleImage(PlanarImage pimgInput, int maxDimension, Interpolation interpolation)
    {
        ParameterBlock paramBlock = new ParameterBlock();
        paramBlock.addSource(pimgInput);
        float xScale = (float)(pimgInput.getWidth());
        float yScale = (float)(pimgInput.getHeight());
        float scaleFactor = (xScale > yScale ? ((1.0f / xScale) * maxDimension) : ((1.0f / yScale) * maxDimension));
        xScale = xScale * scaleFactor;
        yScale = yScale * scaleFactor;
        paramBlock.add(scaleFactor);
        paramBlock.add(scaleFactor);
        paramBlock.add(0.0F);
        paramBlock.add(0.0F);
        if(interpolation == null)
        {
            interpolation = Interpolation.getInstance(Interpolation.INTERP_NEAREST);
        }
        paramBlock.add(interpolation);
        return JAI.create("scale", paramBlock, rHints);
    }

	/** Saves a scaled image to a file
	  */
	public void writeScaledImage(String imageFilename, String saveFilename)
	{
		/* determine Nail Format from menus */
		String nailFormat = "jpeg";
		for(int i = 0; i < jmiNailFormats.length; i++)
		{
			if(jmiNailFormats[i].isSelected())
			{
				nailFormat = ((CustomKeyValuePair)(vcNailFormats.elementAt(i))).getCustomValueAsString(0);
			}
		}
		/* get and write scaled instance of image */
		PlanarImage planarImage = getScaledPlanarImage(imageFilename);
		RenderedImage renderedImage = (RenderedImage)(planarImage.getAsBufferedImage());
		RenderedOp op = JAI.create("filestore", renderedImage, saveFilename + "." + nailFormat, nailFormat);
	}

	/** Saves a thumbnail in the default manner of this application
	  */
	public void writeImageNail(String imageDir, String imageFilename)
	{
		/* determine Nail Format from menus */
		String nailFormat = "jpeg";
		for(int i = 0; i < jmiNailFormats.length; i++)
		{
			if(jmiNailFormats[i].isSelected())
			{
				nailFormat = ((CustomKeyValuePair)(vcNailFormats.elementAt(i))).getCustomValueAsString(0);
			}
		}
		try
		{
			String imageSaveName = getAppPropertyAsString("NAIL_DIR") + File.separator + imageFilename + "." + nailFormat;
			if(jmiReplaceNails.isSelected() || !(new File(imageSaveName)).exists())
			{
				PlanarImage planarImage = getScaledPlanarImage(imageDir + File.separator + imageFilename);
				RenderedImage renderedImage = (RenderedImage)(planarImage.getAsBufferedImage());
				RenderedOp op = JAI.create("filestore", renderedImage, imageSaveName, nailFormat);
			}
			else
			{
				// ignore existing thumbnail image
			}
		}
		catch(Exception e)
		{
			System.out.println("Exception occurred in writeImageNail() on file " + imageFilename);
		}
	}

	/** Writes out a thumbnail page with files grouped by parent directory
	  */
	public void writeTreeFileListing(File outputFile)
	{
		Hashtable htListing = spider.collectListing();
		Enumeration enumKeys = htListing.keys();
		Vector vcDirNames = new Vector();
		Vector[] vcDirContents = new Vector[htListing.size()];
		int counter = 0;
		while(enumKeys.hasMoreElements())
		{
			String key = (String)(enumKeys.nextElement());
			vcDirNames.add(new String(key));
			Vector vcValues = (Vector)(htListing.get(key));
			vcDirContents[counter] = (Vector)(vcValues.clone());
			counter++;
		}
		String nailFormat = "jpeg";
		for(int k = 0; k < jmiNailFormats.length; k++)
		{
			if(jmiNailFormats[k].isSelected()) { nailFormat = ((CustomKeyValuePair)(vcNailFormats.elementAt(k))).getCustomValueAsString(0); }
		}
		try
		{
			FileWriter fw = new FileWriter(outputFile);
			fw.write(getAppPropertyAsString("PAGE_HEADER_BLOCK"), 0, getAppPropertyLength("PAGE_HEADER_BLOCK"));
			if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
			for(int i = 0; i < vcDirNames.size(); i++)
			{
				String key = (String)(vcDirNames.elementAt(i));
				// do not process the nail directory
				if(key.equals(getAppPropertyAsString("NAIL_DIR")))
				{
					continue;
				}
				String currDIR_HEADER_BLOCK = new String(getAppPropertyAsString("DIR_HEADER_BLOCK"));
				if(currDIR_HEADER_BLOCK.indexOf("$TREECOLUMNS$") > -1) { currDIR_HEADER_BLOCK = replaceInString(currDIR_HEADER_BLOCK, "$TREECOLUMNS$", getAppPropertyAsString("TREE_COLUMNS")); }
				if(currDIR_HEADER_BLOCK.indexOf("$DIR$") > -1)         { currDIR_HEADER_BLOCK = replaceInString(currDIR_HEADER_BLOCK, "$DIR$", key); }
				if(currDIR_HEADER_BLOCK.indexOf("$FILESEP$") > -1)     { currDIR_HEADER_BLOCK = replaceInString(currDIR_HEADER_BLOCK, "$FILESEP$", File.separator); }
				fw.write(currDIR_HEADER_BLOCK, 0, currDIR_HEADER_BLOCK.length());
				fw.write(key, 0, key.length());
				fw.write(getAppPropertyAsString("DIR_FOOTER_BLOCK"), 0, getAppPropertyLength("DIR_FOOTER_BLOCK"));
				if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
				int memberCount = 1;
				for(int j = 0; j < vcDirContents[i].size(); j++)
				{
					String value = (String)(vcDirContents[i].elementAt(j));
					if((j == 0) || (memberCount == 1))
					{
						fw.write(getAppPropertyAsString("FIRST_FILE_HEADER_BLOCK"), 0, getAppPropertyLength("FIRST_FILE_HEADER_BLOCK"));
					}
					else
					{
						fw.write(getAppPropertyAsString("FILE_HEADER_BLOCK"), 0, getAppPropertyLength("FILE_HEADER_BLOCK"));
					}
					if(getAppPropertyAsBoolean("LINK_TO_SOURCE"))
					{
						String localLINK_HEADER_BLOCK = new String(getAppPropertyAsString("LINK_HEADER_BLOCK"));
						if(localLINK_HEADER_BLOCK.indexOf("$FILEURL$") > -1)  { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FILEURL$", new File(key + File.separator + value).toURL().toString()); }
						if(localLINK_HEADER_BLOCK.indexOf("$FULLFILE$") > -1) { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FULLFILE$", key + File.separator + value); }
						if(localLINK_HEADER_BLOCK.indexOf("$DIR$") > -1)      { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$DIR$", key); }
						if(localLINK_HEADER_BLOCK.indexOf("$FILESEP$") > -1)  { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FILESEP$", File.separator); }
						if(localLINK_HEADER_BLOCK.indexOf("$FILE$") > -1)     { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FILE$", value); }
						fw.write(localLINK_HEADER_BLOCK, 0, localLINK_HEADER_BLOCK.length());
					}
					setStatusBarText("Processing image " + key + File.separator + value);
					writeImageNail(key, value);
					String localIMAGE_TAG_BLOCK = new String(getAppPropertyAsString("IMAGE_TAG_BLOCK"));
					if(localIMAGE_TAG_BLOCK.indexOf("$NAILURL$") > -1) { localIMAGE_TAG_BLOCK = replaceInString(localIMAGE_TAG_BLOCK, "$NAILURL$", new File(getAppPropertyAsString("NAIL_DIR") + File.separator + value + "." + nailFormat).toURL().toString()); }
					if(localIMAGE_TAG_BLOCK.indexOf("$NAILDIR$") > -1) { localIMAGE_TAG_BLOCK = replaceInString(localIMAGE_TAG_BLOCK, "$NAILDIR$", getAppPropertyAsString("NAIL_DIR")); }
					if(localIMAGE_TAG_BLOCK.indexOf("$FILESEP$") > -1) { localIMAGE_TAG_BLOCK = replaceInString(localIMAGE_TAG_BLOCK, "$FILESEP$", File.separator); }
					if(localIMAGE_TAG_BLOCK.indexOf("$FILE$") > -1)    { localIMAGE_TAG_BLOCK = replaceInString(localIMAGE_TAG_BLOCK, "$FILE$", value + "." + nailFormat); }
					fw.write(localIMAGE_TAG_BLOCK, 0, localIMAGE_TAG_BLOCK.length());
					fw.write(value, 0, value.length());
					if(getAppPropertyAsBoolean("LINK_TO_SOURCE"))
					{
						fw.write(getAppPropertyAsString("LINK_FOOTER_BLOCK"), 0, getAppPropertyLength("LINK_FOOTER_BLOCK"));
					}
					if((j == vcDirContents[i].size() - 1) || (memberCount >= getAppPropertyAsInteger("TREE_COLUMNS")))
					{
						fw.write(getAppPropertyAsString("LAST_FILE_FOOTER_BLOCK"), 0, getAppPropertyLength("LAST_FILE_FOOTER_BLOCK"));
						if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
						memberCount = 1;
					}
					else
					{
						fw.write(getAppPropertyAsString("FILE_FOOTER_BLOCK"), 0, getAppPropertyLength("FILE_FOOTER_BLOCK"));
						memberCount++;
					}
				}
				fw.flush();
			}
			if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
			fw.write(getAppPropertyAsString("PAGE_FOOTER_BLOCK"), 0, getAppPropertyLength("PAGE_FOOTER_BLOCK"));
			fw.close();
			setStatusBarText(MSG_APPREADY);
		}
		catch(IOException ioe)
		{
			sysExit("Error writing tree file : "  + ioe.getMessage(), 3);
		}
	}

	/** Writes out a thumbnail page with files sorted by canonical path and name
	  */
	public void writeIndexFileListing(File outputFile)
	{
		Vector vcBiglist = spider.sortMassListing();
		try
		{
			FileWriter fw = new FileWriter(outputFile);
			fw.write(getAppPropertyAsString("PAGE_HEADER_BLOCK"), 0, getAppPropertyLength("PAGE_HEADER_BLOCK"));
			if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
			int memberCount = 1;
			for(int i = 0; i < vcBiglist.size(); i++)
			{
				String value = (String)(vcBiglist.elementAt(i));
				if((i == 0) || (memberCount == 1))
				{
					fw.write(getAppPropertyAsString("FIRST_FILE_HEADER_BLOCK"), 0, getAppPropertyLength("FIRST_FILE_HEADER_BLOCK"));
				}
				else
				{
					fw.write(getAppPropertyAsString("FILE_HEADER_BLOCK"), 0, getAppPropertyLength("FILE_HEADER_BLOCK"));
				}
				if(getAppPropertyAsBoolean("LINK_TO_SOURCE"))
				{
					String localLINK_HEADER_BLOCK = new String(getAppPropertyAsString("LINK_HEADER_BLOCK"));
					if(localLINK_HEADER_BLOCK.indexOf("$FULLFILE$") > -1) { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FULLFILE$", value); }
					if(localLINK_HEADER_BLOCK.indexOf("$DIR$") > -1)      { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$DIR$", value.substring(0, value.lastIndexOf(File.separator))); }
					if(localLINK_HEADER_BLOCK.indexOf("$FILESEP$") > -1)  { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FILESEP$", File.separator); }
					if(localLINK_HEADER_BLOCK.indexOf("$FILE$") > -1)     { localLINK_HEADER_BLOCK = replaceInString(localLINK_HEADER_BLOCK, "$FILE$", value.substring(value.lastIndexOf(File.separator) + 1, value.length())); }
					fw.write(localLINK_HEADER_BLOCK, 0, localLINK_HEADER_BLOCK.length());
				}
				fw.write(value, 0, value.length());
				if(getAppPropertyAsBoolean("LINK_TO_SOURCE"))
				{
					fw.write(getAppPropertyAsString("LINK_FOOTER_BLOCK"), 0, getAppPropertyLength("LINK_FOOTER_BLOCK"));
				}
				if((i == vcBiglist.size() - 1) || (memberCount >= getAppPropertyAsInteger("INDEX_COLUMNS")))
				{
					fw.write(getAppPropertyAsString("LAST_FILE_FOOTER_BLOCK"), 0, getAppPropertyLength("LAST_FILE_FOOTER_BLOCK"));
					if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
					memberCount = 1;
				}
				else
				{
					fw.write(getAppPropertyAsString("FILE_FOOTER_BLOCK"), 0, getAppPropertyLength("FILE_FOOTER_BLOCK"));
					memberCount++;
				}
				fw.flush();
			}
			if(getAppPropertyAsBoolean("NICE_FORMATTING")) { fw.write("\n\r", 0, 2); }
			fw.write(getAppPropertyAsString("PAGE_FOOTER_BLOCK"), 0, getAppPropertyLength("PAGE_FOOTER_BLOCK"));
			fw.close();
		}
		catch(IOException ioe)
		{
			sysExit("Error writing index file : "  + ioe.getMessage(), 4);
		}
	}

// Application Property Methods ------------------------------------------->

	/* Returns whether an application property exists or not */
	private boolean appPropertyExists(String propertyName)
	{
		return htPropertyTable.containsKey(propertyName);
	}

	/* Retrieve application property as String */
	private String getAppPropertyAsString(String propertyName)
	{
		return (String)(htPropertyTable.get(propertyName));
	}

	/* Retrieve application property as int */
	private int getAppPropertyAsInteger(String propertyName)
	{
		return Integer.parseInt((String)(htPropertyTable.get(propertyName)));
	}

	/* Retrieve application property as boolean */
	private boolean getAppPropertyAsBoolean(String propertyName)
	{
		return ((String)(htPropertyTable.get(propertyName))).equalsIgnoreCase("true");
	}

	/* Retrieve length of application property content String */
	private int getAppPropertyLength(String propertyName)
	{
		return ((String)(htPropertyTable.get(propertyName))).length();
	}

	/* Insert or replace an application property */
	private void updateAppProperty(String propertyName, String newPropertyValue)
	{
		if(htPropertyTable.containsKey(propertyName))
		{
			htPropertyTable.remove(propertyName);
		}
		htPropertyTable.put(propertyName, newPropertyValue);
	}

	/* Replace an application property */
	private void removeAppProperty(String propertyName)
	{
		if(htPropertyTable.containsKey(propertyName))
		{
			htPropertyTable.remove(propertyName);
		}
	}

// Convenience Methods ---------------------------------------------------->

	/* Convenience method for get a file name using a JFileChooser */
	private File getFile(int type, int show, String[] exts, String desc)
	{
		JFileChooser jFileChooser = new JFileChooser(baseDir);
		jFileChooser.setDialogType(type);
		jFileChooser.setFileSelectionMode(show);
		if(exts != null)
		{
			jFileChooser.setFileFilter(new MutableFilter(exts, desc));
		}
		int whichChoice;
		if(type == JFileChooser.SAVE_DIALOG) {
			whichChoice = jFileChooser.showSaveDialog(this);
		}
		else {
			whichChoice = jFileChooser.showOpenDialog(this);
		}
		this.repaint();
		if(whichChoice == JFileChooser.APPROVE_OPTION)
		{
			return jFileChooser.getSelectedFile();
		}
		return null;
	}

	/* Convenience method for setting the title bar text */
	private void setTitleBarText(String newText)
	{
		this.setTitle(APPTITLE + " : " + newText);
	}

	/* Convenience method for setting the status bar text */
	private void setStatusBarText(String newText)
	{
		jlblStatusBar.setText(newText);
	}

	/* Convenience methods for parsing an init file entry */
	private String propFromInit(String initString)
	{
		return initString.substring(0, initString.indexOf("="));
	}

	private String valueFromInit(String initString)
	{
		return initString.substring(initString.indexOf("=") + 1, initString.length());
	}

	/* Convenience method for replacing init file entry variables with values */
	private String replaceInString(String srcString, String findArg, String replArg)
	{
		return srcString.substring(0, srcString.indexOf(findArg)) + replArg + srcString.substring(srcString.indexOf(findArg) + findArg.length(), srcString.length());
	}

	/* Convenience method for exiting on an exception */
	private void sysExit(String msg, int lvl)
	{
		System.out.println(msg);
		System.exit(lvl);
	}

// Main Method ------------------------------------------------------------>

	public static void main(String[] args)
	{
		if(args.length == 4)
		{
			String fdir = "";
			String init = "";
			if(args[0].equals("-d"))      { fdir = args[1]; }
			else if(args[0].equals("-i")) { init = args[1]; }
			else                          { System.out.println("Unknown flag : " + args[0]); System.exit(9); }
			if(args[2].equals("-d"))      { fdir = args[3]; }
			else if(args[2].equals("-i")) { init = args[3]; }
			else                          { System.out.println("Unknown flag : " + args[2]); System.exit(9); }
			Nail6 Nail6 = new Nail6(fdir, init);
		}
		else if(args.length == 2)
		{
			if(args[0].equals("-d"))      { Nail6 Nail6 = new Nail6(args[0]); }
			else if(args[0].equals("-i")) { Nail6 Nail6 = new Nail6(".", args[0]); }
			else                          { System.out.println("Unknown flag : " + args[0]); System.exit(9); }
		}
		else
		{
			Nail6 Nail6 = new Nail6(".");
		}
	}

// Inner Classes ---------------------------------------------------------->

	/* Class for providing JFileChooser with a FileFilter */
	class MutableFilter extends javax.swing.filechooser.FileFilter
	{
		private String[] acceptableExtensions;
		private String descriptor;

		public MutableFilter(String[] exts, String desc)
		{
			acceptableExtensions = exts;
			descriptor           = desc;
		}

    	public boolean accept(File file)
    	{
	        if(file.isDirectory())
			{
				return true;
			}
			String fileName = file.getName();
			if(fileName.indexOf(".") < 0)
			{
				return false;
			}
			String fileExt = fileName.substring(fileName.lastIndexOf("."), fileName.length()).toLowerCase();
			if(fileExt != null)
			{
				for(int i = 0; i < acceptableExtensions.length; i++)
				{
					if(fileExt.equals(acceptableExtensions[i]))
					{
						return true;
					}
				}
				return false;
			}
			else
			{
				return false;
			}
		}

		public String getDescription()
		{
			return descriptor;
		}
	}

	/* Class for menu items that are linked to a specific value */
	class CustomKeyValuePair
	{
		private String   customKey;
		private Object[] customValue;

		public CustomKeyValuePair(String ck, Object[] cv)
		{
			customKey   = ck;
			customValue = cv;
		}

		public String getCustomKey()
		{
			return customKey;
		}

		public Object[] getCustomValue()
		{
			return customValue;
		}

		public Object getCustomValueAsObject(int index)
		{
			return customValue[index];
		}

		public Vector getCustomValueAsVector()
		{
			Vector vcValues = new Vector(customValue.length);
			for(int i = 0; i < customValue.length; i++)
			{
				vcValues.add(customValue[i]);
			}
			return vcValues;
		}

		public int[] getCustomValueAsIntegerArray()
		{
			int[] intValues = new int[customValue.length];
			for(int i = 0; i < customValue.length; i++)
			{
				intValues[i] = Integer.parseInt(customValue[i].toString());
			}
			return intValues;
		}

		public int getCustomValueAsInteger(int index)
		{
			return Integer.parseInt(customValue[index].toString());
		}

		public String[] getCustomValueAsStringArray()
		{
			String[] strValues = new String[customValue.length];
			for(int i = 0; i < customValue.length; i++)
			{
				strValues[i] = customValue[i].toString();
			}
			return strValues;
		}

		public String getCustomValueAsString(int index)
		{
			return customValue[index].toString();
		}
	}

	/* Class for dialog to set application properties */
	class AppOptionsDialog extends JFrame
	{
		public JTextField[] jtxtAppProperties;
		public JCheckBox jchkLINK_TO_SOURCE;
		public JCheckBox jchkNICE_FORMATTING;
		public JCheckBox jchkVIEW_AS_NAIL;
		public JCheckBox jchkOVERWRITENAILS;
		public JCheckBox[] jchkImageTypes;
		public JRadioButton[] jrbtNailSizes;
		public JRadioButton[] jrbtAntiAlias;

		private Font fontText = new Font("Monospaced",Font.PLAIN,12);

		public AppOptionsDialog(Nail6 parent)
		{
			super("Options");
			JPanel jpnlOptionControls = new JPanel();
			JPanel jpnlOptionButtons  = new JPanel();
			JPanel jpnlImageTypes     = new JPanel();
			JPanel jpnlNailSize       = new JPanel();
			JPanel jpnlAntiAlias      = new JPanel();

			JButton jbtnSave        = new JButton("Save");   jbtnSave.setActionCommand("saveoptions");              jbtnSave.addActionListener(parent);
			JButton jbtnCancel      = new JButton("Cancel"); jbtnCancel.setActionCommand("closeoptions");           jbtnCancel.addActionListener(parent);
			JButton jbtnBrowseImage = new JButton("Browse"); jbtnBrowseImage.setActionCommand("browseforimagedir"); jbtnBrowseImage.addActionListener(parent);
			JButton jbtnBrowseNail  = new JButton("Browse"); jbtnBrowseNail.setActionCommand("browsefornaildir");   jbtnBrowseNail.addActionListener(parent);

			jchkLINK_TO_SOURCE   = new JCheckBox("", Nail6.this.getAppPropertyAsBoolean("LINK_TO_SOURCE"));
			jchkNICE_FORMATTING = new JCheckBox("", Nail6.this.getAppPropertyAsBoolean("NICE_FORMATTING"));
			jchkVIEW_AS_NAIL     = new JCheckBox("", Nail6.this.getAppPropertyAsBoolean("VIEW_AS_NAIL"));

			jpnlImageTypes = new JPanel();
			jpnlImageTypes.setLayout(new GridLayout(1, Nail6.this.vcImageTypes.size()));
			jchkImageTypes = new JCheckBox[Nail6.this.vcImageTypes.size()];
			for(int i = 0; i < Nail6.this.vcImageTypes.size(); i++)
			{
				String ckvpImageType = ((CustomKeyValuePair)(Nail6.this.vcImageTypes.elementAt(i))).getCustomKey();
				jchkImageTypes[i] = new JCheckBox(ckvpImageType, appPropertyExists(ARRAY_IMAGETYPE + ckvpImageType));
				jpnlImageTypes.add(jchkImageTypes[i]);
			}

			jpnlNailSize = new JPanel();
			jpnlNailSize.setLayout(new GridLayout(1, Nail6.this.vcNailSizes.size()));
			ButtonGroup btngNailSizes = new ButtonGroup();
			jrbtNailSizes = new JRadioButton[Nail6.this.vcNailSizes.size()];
			for(int i = 0; i < Nail6.this.vcNailSizes.size(); i++)
			{
				CustomKeyValuePair ckvpNailSize = (CustomKeyValuePair)(Nail6.this.vcNailSizes.elementAt(i));
				jrbtNailSizes[i] = new JRadioButton(ckvpNailSize.getCustomKey(), ckvpNailSize.getCustomValueAsInteger(0) == getAppPropertyAsInteger("DEFAULT_NAIL_SIZE"));
				btngNailSizes.add(jrbtNailSizes[i]);
				jpnlNailSize.add(jrbtNailSizes[i]);
			}

			jpnlAntiAlias = new JPanel();
			jpnlAntiAlias.setLayout(new GridLayout(1, Nail6.this.vcInterpTypes.size()));
			ButtonGroup btngAntiAlias = new ButtonGroup();
			jrbtAntiAlias = new JRadioButton[Nail6.this.vcInterpTypes.size()];
			for(int i = 0; i < Nail6.this.vcInterpTypes.size(); i++)
			{
				String ckvpAntiAlias = ((CustomKeyValuePair)(Nail6.this.vcInterpTypes.elementAt(i))).getCustomKey();
				jrbtAntiAlias[i] = new JRadioButton(ckvpAntiAlias, ckvpAntiAlias.equals(getAppPropertyAsString("DEFAULT_ANTIALIAS")));
				btngAntiAlias.add(jrbtAntiAlias[i]);
				jpnlAntiAlias.add(jrbtAntiAlias[i]);
			}

			int irow = 0;
			jpnlOptionControls.setLayout(new GridBagLayout());
			GridBagConstraints gbc = new GridBagConstraints();
			gbc.gridheight = 1; gbc.gridwidth = 1; gbc.weighty = 1; gbc.ipadx = 2; gbc.ipady = 2; gbc.anchor = GridBagConstraints.WEST; gbc.fill = GridBagConstraints.HORIZONTAL;
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(new JLabel("APPLICATION PROPERTIES"), gbc);

			jtxtAppProperties = new JTextField[Nail6.this.vcPropertySequence.size()];
			for(int i = 0; i < Nail6.this.vcPropertySequence.size(); i++)
			{
				String appPropName = (String)(Nail6.this.vcPropertySequence.elementAt(i));
				if(appPropName.equals("DEFAULT_NAIL_SIZE") || appPropName.equals("DEFAULT_ANTIALIAS")) { continue; } // handled by radio buttons lower down
				jtxtAppProperties[i] = new JTextField(Nail6.this.getAppPropertyAsString(appPropName), 64);
				jtxtAppProperties[i].setFont(fontText);
				gbc.gridx = 0; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(new JLabel(appPropName), gbc);
				if(appPropName.equals("DEFAULT_IMAGE_DIR"))
				{
					gbc.gridx = 1; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 1; jpnlOptionControls.add(jtxtAppProperties[i], gbc);
					gbc.gridx = 2; gbc.gridy = irow++; gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(jbtnBrowseImage, gbc);
				}
				else if(appPropName.equals("NAIL_DIR"))
				{
					gbc.gridx = 1; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 1; jpnlOptionControls.add(jtxtAppProperties[i], gbc);
					gbc.gridx = 2; gbc.gridy = irow++; gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(jbtnBrowseNail, gbc);
				}
				else if(appPropName.equals("LINK_TO_SOURCE"))
				{
					gbc.gridx = 1; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 1; jpnlOptionControls.add(jchkLINK_TO_SOURCE, gbc);
					gbc.gridx = 2; gbc.gridy = irow++; gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(new JLabel(""), gbc);
				}
				else if(appPropName.equals("NICE_FORMATTING"))
				{
					gbc.gridx = 1; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 1; jpnlOptionControls.add(jchkNICE_FORMATTING, gbc);
					gbc.gridx = 2; gbc.gridy = irow++; gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(new JLabel(""), gbc);
				}
				else if(appPropName.equals("VIEW_AS_NAIL"))
				{
					gbc.gridx = 1; gbc.gridy = irow;   gbc.gridwidth = 1; gbc.weightx = 1; jpnlOptionControls.add(jchkVIEW_AS_NAIL, gbc);
					gbc.gridx = 2; gbc.gridy = irow++; gbc.gridwidth = 1; gbc.weightx = 0; jpnlOptionControls.add(new JLabel(""), gbc);
				}
				else
				{
					gbc.gridx = 1; gbc.gridy = irow++; gbc.gridwidth = 2; gbc.weightx = 1; jpnlOptionControls.add(jtxtAppProperties[i], gbc);
				}
			}
			gbc.fill = GridBagConstraints.NONE;
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(new JLabel("IMAGE TYPES"), gbc);
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(jpnlImageTypes, gbc);
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(new JLabel("DEFAULT NAIL SIZE"), gbc);
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(jpnlNailSize, gbc);
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(new JLabel("DEFAULT ANTIALIASING"), gbc);
			gbc.gridx = 0; gbc.gridy = irow++; gbc.gridwidth = 3; gbc.weightx = 1; jpnlOptionControls.add(jpnlAntiAlias, gbc);

			jpnlOptionButtons.setLayout(new GridLayout(1, 2));
			jpnlOptionButtons.add(jbtnSave);
			jpnlOptionButtons.add(jbtnCancel);

			this.getContentPane().setLayout(new BorderLayout());
			this.getContentPane().add(jpnlOptionControls, BorderLayout.CENTER);
			this.getContentPane().add(jpnlOptionButtons, BorderLayout.SOUTH);
			this.pack();
			this.show();
		}
	}

}