/*
GNU General Public License

NailIcon
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.nail6;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.awt.image.renderable.ParameterBlock;
import java.awt.geom.AffineTransform;
import javax.media.jai.InterpolationNearest;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.swing.Icon;

/** NailIcon
  * Icon image class for Nail 6
  *
  * @author Howard Kistler
  * @version 0.1
  *
  * REQUIREMENTS
  * Java 2 (JDK 1.3)
  * Swing
  * Java Advanced Imaging library
  *
  * VERSION HISTORY
  * 0.1 (09/14/2000) - initial creation (started on 08/09/2000)
  *
  * ACKNOWLEDGMENT
  * Derived from Sun code in Java Advanced Imaging demo
  */

public class NailIcon implements Icon
{

    protected RenderedImage riSource;
    protected SampleModel   sampleModel;
    protected ColorModel    colorModel;

    protected int minX;
    protected int minY;
    protected int width;
    protected int height;
    protected Rectangle imageBounds;

    protected int minTileX;
    protected int maxTileX;
    protected int minTileY;
    protected int maxTileY;
    protected int tileWidth;
    protected int tileHeight;
    protected int tileGridXOffset;
    protected int tileGridYOffset;
    protected int originX;
    protected int originY;

    private Color defaultBackgroundColor = null;
    private Color backgroundColor = null;

    public NailIcon(RenderedImage riSource, Color dbgc)
    {
    	defaultBackgroundColor = dbgc;

        int mx = riSource.getMinX();
        int my = riSource.getMinY();
        if ((mx < 0) || (my < 0))
        {
            ParameterBlock pb = new ParameterBlock();
            pb.addSource(riSource);
            pb.add((float)Math.max(-mx, 0));
            pb.add((float)Math.max(-my, 0));
            pb.add(new InterpolationNearest());
            riSource = JAI.create("translate", pb, null);
        }
	
        this.riSource = riSource;
        this.sampleModel = riSource.getSampleModel();
        this.colorModel = riSource.getColorModel();
        if (this.colorModel == null)
        {
            this.colorModel = PlanarImage.createColorModel(riSource.getSampleModel());
        }
        if (this.colorModel == null)
        {
            throw new IllegalArgumentException("NailIcon is unable to display supplied RenderedImage.");
        }

        if (this.colorModel.getTransparency() != Transparency.OPAQUE)
        {
            Object col = riSource.getProperty("background_color");
            if (col != null)
            {
                backgroundColor = (Color)col;
            }
            else
            {
                backgroundColor = defaultBackgroundColor;
            }
        }

        minX = riSource.getMinX();
        minY = riSource.getMinY();
        width = riSource.getWidth();
        height = riSource.getHeight();
        imageBounds = new Rectangle(minX, minY, width, height);

        minTileX = riSource.getMinTileX();
        maxTileX = riSource.getMinTileX() + riSource.getNumXTiles() - 1;
        minTileY = riSource.getMinTileY();
        maxTileY = riSource.getMinTileY() + riSource.getNumYTiles() - 1;
        tileWidth  = riSource.getTileWidth();
        tileHeight = riSource.getTileHeight();
        tileGridXOffset = riSource.getTileGridXOffset();
        tileGridYOffset = riSource.getTileGridYOffset();

        originX = originY = 0;
    }

    public NailIcon(RenderedImage riSource)
	{
		this(riSource, new Color(128, 128, 128));
	}

    public int getIconWidth()
    {
        return riSource.getWidth();
    }

    public int getIconHeight()
    {
        return riSource.getHeight();
    }

    public void paintIcon(Component c, Graphics g, int x, int y)
    {
        Graphics2D g2D = null;
        if (g instanceof Graphics2D)
        {
            g2D = (Graphics2D)g;
        }
        else
        {
            System.err.println("NailIcon: not a Graphics2D.");
            return;
        }

        // Get the clipping rectangle and translate it into image coordinates. 
        Rectangle clipBounds = g.getClipBounds();

        int transX = x;
        int transY = y;

        // Determine the extent of the clipping region in tile coordinates.
        int txmin, txmax, tymin, tymax;
        int ti, tj;

        // Constrain drawing to the active image area
        Rectangle imageRect = new Rectangle(minX + transX, minY + transY, width, height);
        if (clipBounds == null)
        {
     	    clipBounds = new Rectangle(0,0,c.getWidth(), c.getHeight());
     	}
	  
		//txmin = XtoTileX(clipBounds.x);
		txmin = PlanarImage.XToTileX(clipBounds.x - transX, tileGridXOffset, tileWidth);
		txmin = Math.max(txmin, minTileX);
		txmin = Math.min(txmin, maxTileX);
	
		//txmax = XtoTileX(clipBounds.x + clipBounds.width - 1);
		txmax = PlanarImage.XToTileX(clipBounds.x + clipBounds.width - 1 - transX, tileGridXOffset, tileWidth );
		txmax = Math.max(txmax, minTileX);
		txmax = Math.min(txmax, maxTileX);
	
		//tymin = YtoTileY(clipBounds.y);
		tymin = PlanarImage.YToTileY(clipBounds.y - transY, tileGridYOffset, tileHeight);
		tymin = Math.max(tymin, minTileY);
		tymin = Math.min(tymin, maxTileY);
	
		//tymax = YtoTileY(clipBounds.y + clipBounds.height - 1);
		tymax = PlanarImage.YToTileY(clipBounds.y + clipBounds.height - 1 - transY, tileGridYOffset, tileHeight);
		tymax = Math.max(tymax, minTileY);
		tymax = Math.min(tymax, maxTileY);
	
		g2D.clip(imageRect);
	
        if (backgroundColor != null)
        {
            g2D.setColor(backgroundColor);
        }

        // Loop over tiles within the clipping region
        for (tj = tymin; tj <= tymax; tj++)
        {
            for (ti = txmin; ti <= txmax; ti++)
            {
                int tx = PlanarImage.tileXToX(ti, tileGridXOffset, tileWidth);
                int ty = PlanarImage.tileYToY(tj, tileGridYOffset, tileHeight);

                Raster tile = riSource.getTile(ti, tj);
                DataBuffer dataBuffer = tile.getDataBuffer();

                Point origin = new Point(0, 0);
                WritableRaster wr = tile.createWritableRaster(sampleModel, dataBuffer, origin);

                BufferedImage bi = new BufferedImage(colorModel, wr, false, null);

                AffineTransform transform = AffineTransform.getTranslateInstance(tx + transX, ty + transY);

                if (backgroundColor != null)
                {
                    g2D.fillRect(tx + transX, ty + transY, tileWidth, tileHeight);
                }
                g2D.drawImage(bi, transform, null);
            }
        }
    }
}
